@@ -27,7 +27,57 @@ handler.DELAY = 0.1
 swallow_http_errors = handler.swallow_http_errors
 
 
-class TestAppBasic(AsyncHTTPTestCase):
+class TestAppBase(AsyncHTTPTestCase):
+
+    def get_httpserver_options(self):
+        return get_server_settings(options)
+
+    def assert_response(self, bstr, response):
+        if swallow_http_errors:
+            self.assertEqual(response.code, 200)
+            self.assertIn(bstr, response.body)
+        else:
+            self.assertEqual(response.code, 400)
+            self.assertIn(b'Bad Request', response.body)
+
+    def assert_status_in(self, data, status):
+        self.assertIsNone(data['encoding'])
+        self.assertIsNone(data['id'])
+        self.assertIn(status, data['status'])
+
+    def assert_status_equal(self, data, status):
+        self.assertIsNone(data['encoding'])
+        self.assertIsNone(data['id'])
+        self.assertEqual(status, data['status'])
+
+    def assert_status_none(self, data):
+        self.assertIsNotNone(data['encoding'])
+        self.assertIsNotNone(data['id'])
+        self.assertIsNone(data['status'])
+
+    def fetch_request(self, url, method='GET', body='', headers={}, sync=True):
+        if not sync and url.startswith('/'):
+            url = self.get_url(url)
+
+        if isinstance(body, dict):
+            body = urlencode(body)
+
+        if not headers:
+            headers = self.headers
+        else:
+            headers.update(self.headers)
+
+        client = self if sync else self.get_http_client()
+        return client.fetch(url, method=method, body=body, headers=headers)
+
+    def sync_post(self, url, body, headers={}):
+        return self.fetch_request(url, 'POST', body, headers)
+
+    def async_post(self, url, body, headers={}):
+        return self.fetch_request(url, 'POST', body, headers, sync=False)
+
+
+class TestAppBasic(TestAppBase):
 
     running = [True]
     sshserver_port = 2200
@@ -65,94 +115,68 @@ class TestAppBasic(AsyncHTTPTestCase):
         cls.running.pop()
         print('='*20)
 
-    def get_httpserver_options(self):
-        return get_server_settings(options)
-
-    def assert_response(self, bstr, response):
-        if swallow_http_errors:
-            self.assertEqual(response.code, 200)
-            self.assertIn(bstr, response.body)
-        else:
-            self.assertEqual(response.code, 400)
-            self.assertIn(b'Bad Request', response.body)
-
-    def sync_post(self, body, headers={}, url='/', method='POST'):
-        headers.update(self.headers)
-        return self.fetch(url, method=method, body=body, headers=headers)
-
-    def async_post(self, url, body, headers={}, method='POST'):
-        client = self.get_http_client()
-        headers.update(self.headers)
-        return client.fetch(url, method=method, body=body, headers=headers)
-
     def test_app_with_invalid_form_for_missing_argument(self):
         response = self.fetch('/')
         self.assertEqual(response.code, 200)
 
         body = 'port=7000&username=admin&password&_xsrf=yummy'
-        response = self.sync_post(body)
+        response = self.sync_post('/', body)
         self.assert_response(b'Missing argument hostname', response)
 
         body = 'hostname=127.0.0.1&port=7000&password&_xsrf=yummy'
-        response = self.sync_post(body)
+        response = self.sync_post('/', body)
         self.assert_response(b'Missing argument username', response)
 
         body = 'hostname=&port=&username=&password&_xsrf=yummy'
-        response = self.sync_post(body)
+        response = self.sync_post('/', body)
         self.assert_response(b'Missing value hostname', response)
 
         body = 'hostname=127.0.0.1&port=7000&username=&password&_xsrf=yummy'
-        response = self.sync_post(body)
+        response = self.sync_post('/', body)
         self.assert_response(b'Missing value username', response)
 
     def test_app_with_invalid_form_for_invalid_value(self):
         body = 'hostname=127.0.0&port=22&username=&password&_xsrf=yummy'
-        response = self.sync_post(body)
+        response = self.sync_post('/', body)
         self.assert_response(b'Invalid hostname', response)
 
         body = 'hostname=http://www.googe.com&port=22&username=&password&_xsrf=yummy'  # noqa
-        response = self.sync_post(body)
+        response = self.sync_post('/', body)
         self.assert_response(b'Invalid hostname', response)
 
         body = 'hostname=127.0.0.1&port=port&username=&password&_xsrf=yummy'
-        response = self.sync_post(body)
+        response = self.sync_post('/', body)
         self.assert_response(b'Invalid port', response)
 
         body = 'hostname=127.0.0.1&port=70000&username=&password&_xsrf=yummy'
-        response = self.sync_post(body)
+        response = self.sync_post('/', body)
         self.assert_response(b'Invalid port', response)
 
     def test_app_with_wrong_hostname_ip(self):
         body = 'hostname=127.0.0.2&port=2200&username=admin&_xsrf=yummy'
-        response = self.sync_post(body)
+        response = self.sync_post('/', body)
         self.assertEqual(response.code, 200)
         self.assertIn(b'Unable to connect to', response.body)
 
     def test_app_with_wrong_hostname_domain(self):
         body = 'hostname=xxxxxxxxxxxx&port=2200&username=admin&_xsrf=yummy'
-        response = self.sync_post(body)
+        response = self.sync_post('/', body)
         self.assertEqual(response.code, 200)
         self.assertIn(b'Unable to connect to', response.body)
 
     def test_app_with_wrong_port(self):
         body = 'hostname=127.0.0.1&port=7000&username=admin&_xsrf=yummy'
-        response = self.sync_post(body)
+        response = self.sync_post('/', body)
         self.assertEqual(response.code, 200)
         self.assertIn(b'Unable to connect to', response.body)
 
     def test_app_with_wrong_credentials(self):
-        response = self.sync_post(self.body + 's')
-        data = json.loads(to_str(response.body))
-        self.assertIsNone(data['encoding'])
-        self.assertIsNone(data['id'])
-        self.assertIn('Authentication failed.', data['status'])
+        response = self.sync_post('/', self.body + 's')
+        self.assert_status_in(json.loads(to_str(response.body)), 'Authentication failed.') # noqa
 
     def test_app_with_correct_credentials(self):
-        response = self.sync_post(self.body)
-        data = json.loads(to_str(response.body))
-        self.assertIsNone(data['status'])
-        self.assertIsNotNone(data['id'])
-        self.assertIsNotNone(data['encoding'])
+        response = self.sync_post('/', self.body)
+        self.assert_status_none(json.loads(to_str(response.body)))
 
     def test_app_with_correct_credentials_but_with_no_port(self):
         default_port = handler.DEFAULT_PORT
@@ -160,19 +184,13 @@ class TestAppBasic(AsyncHTTPTestCase):
 
         # with no port value
         body = self.body.replace(str(self.sshserver_port), '')
-        response = self.sync_post(body)
-        data = json.loads(to_str(response.body))
-        self.assertIsNone(data['status'])
-        self.assertIsNotNone(data['id'])
-        self.assertIsNotNone(data['encoding'])
+        response = self.sync_post('/', body)
+        self.assert_status_none(json.loads(to_str(response.body)))
 
         # with no port argument
         body = body.replace('port=&', '')
-        response = self.sync_post(body)
-        data = json.loads(to_str(response.body))
-        self.assertIsNone(data['status'])
-        self.assertIsNotNone(data['id'])
-        self.assertIsNotNone(data['encoding'])
+        response = self.sync_post('/', body)
+        self.assert_status_none(json.loads(to_str(response.body)))
 
         handler.DEFAULT_PORT = default_port
 
@@ -181,9 +199,7 @@ class TestAppBasic(AsyncHTTPTestCase):
         url = self.get_url('/')
         response = yield self.async_post(url, self.body)
         data = json.loads(to_str(response.body))
-        self.assertIsNone(data['status'])
-        self.assertIsNotNone(data['id'])
-        self.assertIsNotNone(data['encoding'])
+        self.assert_status_none(data)
 
         url = url.replace('http', 'ws')
         ws_url = url + 'ws?id=' + data['id']
@@ -198,9 +214,7 @@ class TestAppBasic(AsyncHTTPTestCase):
         url = self.get_url('/')
         response = yield self.async_post(url, self.body)
         data = json.loads(to_str(response.body))
-        self.assertIsNone(data['status'])
-        self.assertIsNotNone(data['id'])
-        self.assertIsNotNone(data['encoding'])
+        self.assert_status_none(data)
 
         url = url.replace('http', 'ws')
         ws_url = url + 'ws?id=' + data['id']
@@ -214,9 +228,7 @@ class TestAppBasic(AsyncHTTPTestCase):
         url = self.get_url('/')
         response = yield self.async_post(url, self.body)
         data = json.loads(to_str(response.body))
-        self.assertIsNone(data['status'])
-        self.assertIsNotNone(data['id'])
-        self.assertIsNotNone(data['encoding'])
+        self.assert_status_none(data)
 
         url = url.replace('http', 'ws')
         ws_url = url + 'ws'
@@ -230,9 +242,7 @@ class TestAppBasic(AsyncHTTPTestCase):
         url = self.get_url('/')
         response = yield self.async_post(url, self.body)
         data = json.loads(to_str(response.body))
-        self.assertIsNone(data['status'])
-        self.assertIsNotNone(data['id'])
-        self.assertIsNotNone(data['encoding'])
+        self.assert_status_none(data)
 
         url = url.replace('http', 'ws')
         ws_url = url + 'ws?id='
@@ -246,9 +256,7 @@ class TestAppBasic(AsyncHTTPTestCase):
         url = self.get_url('/')
         response = yield self.async_post(url, self.body)
         data = json.loads(to_str(response.body))
-        self.assertIsNone(data['status'])
-        self.assertIsNotNone(data['id'])
-        self.assertIsNotNone(data['encoding'])
+        self.assert_status_none(data)
 
         url = url.replace('http', 'ws')
         ws_url = url + 'ws?id=1' + data['id']
@@ -263,9 +271,7 @@ class TestAppBasic(AsyncHTTPTestCase):
         url = self.get_url('/')
         response = yield self.async_post(url, body)
         data = json.loads(to_str(response.body))
-        self.assertIsNone(data['status'])
-        self.assertIsNotNone(data['id'])
-        self.assertIsNotNone(data['encoding'])
+        self.assert_status_none(data)
 
         url = url.replace('http', 'ws')
         ws_url = url + 'ws?id=' + data['id']
@@ -307,12 +313,9 @@ class TestAppBasic(AsyncHTTPTestCase):
         url = self.get_url('/')
         privatekey = read_file(make_tests_data_path('user_rsa_key'))
         self.body_dict.update(privatekey=privatekey)
-        body = urlencode(self.body_dict)
-        response = yield self.async_post(url, body)
+        response = yield self.async_post(url, self.body_dict)
         data = json.loads(to_str(response.body))
-        self.assertIsNone(data['status'])
-        self.assertIsNotNone(data['id'])
-        self.assertIsNotNone(data['encoding'])
+        self.assert_status_none(data)
 
         url = url.replace('http', 'ws')
         ws_url = url + 'ws?id=' + data['id']
@@ -333,9 +336,7 @@ class TestAppBasic(AsyncHTTPTestCase):
         }
         response = yield self.async_post(url, body, headers=headers)
         data = json.loads(to_str(response.body))
-        self.assertIsNone(data['status'])
-        self.assertIsNotNone(data['id'])
-        self.assertIsNotNone(data['encoding'])
+        self.assert_status_none(data)
 
         url = url.replace('http', 'ws')
         ws_url = url + 'ws?id=' + data['id']
@@ -402,9 +403,7 @@ class TestAppBasic(AsyncHTTPTestCase):
                 yield self.async_post(url, body, headers=headers)
             self.assertIn('Bad Request', ctx.exception.message)
 
-    @tornado.testing.gen_test
     def test_app_post_form_with_large_body_size_by_multipart_form(self):
-        url = self.get_url('/')
         privatekey = 'h' * (2 * max_body_size)
         files = [('privatekey', 'user_rsa_key', privatekey)]
         content_type, body = encode_multipart_formdata(self.body_dict.items(),
@@ -412,34 +411,24 @@ class TestAppBasic(AsyncHTTPTestCase):
         headers = {
             'Content-Type': content_type, 'content-length': str(len(body))
         }
+        response = self.sync_post('/', body, headers=headers)
+        self.assertIn(response.code, [400, 599])
 
-        with self.assertRaises(HTTPError) as ctx:
-            yield self.async_post(url, body, headers=headers)
-        self.assertIn('Bad Request', ctx.exception.message)
-
-    @tornado.testing.gen_test
     def test_app_post_form_with_large_body_size_by_urlencoded_form(self):
-        url = self.get_url('/')
         privatekey = 'h' * (2 * max_body_size)
         body = self.body + '&privatekey=' + privatekey
-        with self.assertRaises(HTTPError) as ctx:
-            yield self.async_post(url, body)
-        self.assertIn('Bad Request', ctx.exception.message)
+        response = self.sync_post('/', body)
+        self.assertIn(response.code, [400, 599])
 
     @tornado.testing.gen_test
     def test_app_with_user_keyonly_for_bad_authentication_type(self):
-        url = self.get_url('/')
         self.body_dict.update(username='keyonly', password='foo')
-        body = urlencode(self.body_dict)
-        response = yield self.async_post(url, body)
+        response = yield self.async_post('/', self.body_dict)
         self.assertEqual(response.code, 200)
-        data = json.loads(to_str(response.body))
-        self.assertIsNone(data['id'])
-        self.assertIsNone(data['encoding'])
-        self.assertIn('Bad authentication type', data['status'])
+        self.assert_status_in(json.loads(to_str(response.body)), 'Bad authentication type') # noqa
 
 
-class OtherTestBase(AsyncHTTPTestCase):
+class OtherTestBase(TestAppBase):
     sshserver_port = 3300
     headers = {'Cookie': '_xsrf=yummy'}
     debug = False
@@ -472,9 +461,6 @@ class OtherTestBase(AsyncHTTPTestCase):
         app = make_app(make_handlers(loop, options), get_app_settings(options))
         return app
 
-    def get_httpserver_options(self):
-        return get_server_settings(options)
-
     def setUp(self):
         print('='*20)
         self.running = True
@@ -506,12 +492,8 @@ class TestAppInDebugMode(OtherTestBase):
             self.assertIn(b'Uncaught exception', response.body)
 
     def test_server_error_for_post_method(self):
-        response = self.fetch(
-            '/',
-            method='POST',
-            body=urlencode(dict(self.body, error='raise')),
-            headers=self.headers
-        )
+        body = dict(self.body, error='raise')
+        response = self.sync_post('/', body)
         self.assert_response(b'"status": "Internal Server Error"', response)
 
     def test_html(self):
@@ -524,14 +506,9 @@ class TestAppWithLargeBuffer(OtherTestBase):
     @tornado.testing.gen_test
     def test_app_for_sending_message_with_large_size(self):
         url = self.get_url('/')
-        client = self.get_http_client()
-        body = urlencode(dict(self.body, username='foo'))
-        response = yield client.fetch(url, method='POST', body=body,
-                                      headers=self.headers)
+        response = yield self.async_post(url, dict(self.body, username='foo'))
         data = json.loads(to_str(response.body))
-        self.assertIsNone(data['status'])
-        self.assertIsNotNone(data['id'])
-        self.assertIsNotNone(data['encoding'])
+        self.assert_status_none(data)
 
         url = url.replace('http', 'ws')
         ws_url = url + 'ws?id=' + data['id']
@@ -559,14 +536,8 @@ class TestAppWithRejectPolicy(OtherTestBase):
 
     @tornado.testing.gen_test
     def test_app_with_hostname_not_in_hostkeys(self):
-        url = self.get_url('/')
-        client = self.get_http_client()
-        body = urlencode(dict(self.body, username='foo'))
-        response = yield client.fetch(url, method='POST', body=body,
-                                      headers=self.headers)
+        response = yield self.async_post('/', self.body)
         data = json.loads(to_str(response.body))
-        self.assertIsNone(data['id'])
-        self.assertIsNone(data['encoding'])
         message = 'Connection to {}:{} is not allowed.'.format(self.body['hostname'], self.sshserver_port) # noqa
         self.assertEqual(message, data['status'])
 
@@ -582,14 +553,8 @@ class TestAppWithBadHostKey(OtherTestBase):
 
     @tornado.testing.gen_test
     def test_app_with_bad_host_key(self):
-        url = self.get_url('/')
-        client = self.get_http_client()
-        body = urlencode(dict(self.body, username='foo'))
-        response = yield client.fetch(url, method='POST', body=body,
-                                      headers=self.headers)
+        response = yield self.async_post('/', self.body)
         data = json.loads(to_str(response.body))
-        self.assertIsNone(data['id'])
-        self.assertIsNone(data['encoding'])
         self.assertEqual('Bad host key.', data['status'])
 
 
@@ -602,14 +567,12 @@ class TestAppWithTrustedStream(OtherTestBase):
         self.assertIn('Forbidden', response.error.message)
 
     def test_with_forbidden_post_request(self):
-        response = self.fetch('/', method='POST', body=urlencode(self.body),
-                              headers=self.headers)
+        response = self.sync_post('/', self.body)
         self.assertEqual(response.code, 403)
         self.assertIn('Forbidden', response.error.message)
 
     def test_with_forbidden_put_request(self):
-        response = self.fetch('/', method='PUT', body=urlencode(self.body),
-                              headers=self.headers)
+        response = self.fetch_request('/', method='PUT', body=self.body)
         self.assertEqual(response.code, 403)
         self.assertIn('Forbidden', response.error.message)
 
@@ -627,8 +590,7 @@ class TestAppNotFoundHandler(OtherTestBase):
         self.assertIn(b'404: Not Found', response.body)
 
     def test_with_not_found_post_request(self):
-        response = self.fetch('/pathnotfound', method='POST',
-                              body=urlencode(self.body), headers=self.headers)
+        response = self.sync_post('/pathnotfound', self.body)
         self.assertEqual(response.code, 404)
         self.assertEqual(
             response.headers['Server'], self.custom_headers['Server']
@@ -636,8 +598,8 @@ class TestAppNotFoundHandler(OtherTestBase):
         self.assertIn(b'404: Not Found', response.body)
 
     def test_with_not_found_put_request(self):
-        response = self.fetch('/pathnotfound', method='PUT',
-                              body=urlencode(self.body), headers=self.headers)
+        response = self.fetch_request('/pathnotfound', method='PUT',
+                                      body=self.body)
         self.assertEqual(response.code, 404)
         self.assertEqual(
             response.headers['Server'], self.custom_headers['Server']
@@ -666,14 +628,8 @@ class TestAppWithPutRequest(OtherTestBase):
 
     @tornado.testing.gen_test
     def test_app_with_method_not_supported(self):
-        url = self.get_url('/')
-        client = self.get_http_client()
-        body = urlencode(dict(self.body, username='foo'))
-
         with self.assertRaises(HTTPError) as ctx:
-            yield client.fetch(
-                url, method='PUT', body=body, headers=self.headers
-            )
+            yield self.fetch_request('/', 'PUT', self.body, sync=False)
         self.assertIn('Method Not Allowed', ctx.exception.message)
 
 
@@ -687,59 +643,46 @@ class TestAppWithTooManyConnections(OtherTestBase):
 
     @tornado.testing.gen_test
     def test_app_with_too_many_connections(self):
+        clients['127.0.0.1'] = {'fake_worker_id': None}
+
         url = self.get_url('/')
-        client = self.get_http_client()
-        body = urlencode(dict(self.body, username='foo'))
-        response = yield client.fetch(url, method='POST', body=body,
-                                      headers=self.headers)
+        response = yield self.async_post(url, self.body)
         data = json.loads(to_str(response.body))
-        worker_id = data['id']
-        self.assertIsNotNone(worker_id)
-        self.assertIsNotNone(data['encoding'])
-        self.assertIsNone(data['status'])
+        self.assertEqual('Too many live connections.', data['status'])
 
-        response = yield client.fetch(url, method='POST', body=body,
-                                      headers=self.headers)
-        data = json.loads(to_str(response.body))
-        self.assertIsNone(data['id'])
-        self.assertIsNone(data['encoding'])
-        self.assertEqual(data['status'], 'Too many connections.')
-
-        ws_url = url.replace('http', 'ws') + 'ws?id=' + worker_id
-        ws = yield tornado.websocket.websocket_connect(ws_url)
-        msg = yield ws.read_message()
-        self.assertIsNotNone(msg)
-
-        response = yield client.fetch(url, method='POST', body=body,
-                                      headers=self.headers)
-        data = json.loads(to_str(response.body))
-        self.assertIsNone(data['id'])
-        self.assertIsNone(data['encoding'])
-        self.assertEqual(data['status'], 'Too many connections.')
-
-        ws.close()
+        clients['127.0.0.1'].clear()
+        response = yield self.async_post(url, self.body)
+        self.assert_status_none(json.loads(to_str(response.body)))
 
 
-class TestAppWithCrossOriginConnect(OtherTestBase):
+class TestAppWithCrossOriginOperation(OtherTestBase):
 
     origin = 'http://www.example.com'
 
     @tornado.testing.gen_test
-    def test_app_with_cross_orgin_connect(self):
-        url = self.get_url('/')
-        client = self.get_http_client()
-        body = urlencode(dict(self.body, username='foo', _origin='localhost'))
-        response = yield client.fetch(url, method='POST', body=body,
-                                      headers=self.headers)
-        data = json.loads(to_str(response.body))
-        self.assertIsNone(data['id'])
-        self.assertIsNone(data['encoding'])
-        self.assertIn('Cross origin frame', data['status'])
+    def test_app_with_wrong_event_origin(self):
+        body = dict(self.body, _origin='localhost')
+        response = yield self.async_post('/', body)
+        self.assert_status_equal(json.loads(to_str(response.body)), 'Cross origin operation is not allowed.') # noqa
 
-        body = urlencode(dict(self.body, username='foo', _origin=self.origin))
-        response = yield client.fetch(url, method='POST', body=body,
-                                      headers=self.headers)
-        data = json.loads(to_str(response.body))
-        self.assertIsNotNone(data['id'])
-        self.assertIsNotNone(data['encoding'])
-        self.assertIsNone(data['status'])
+    @tornado.testing.gen_test
+    def test_app_with_wrong_header_origin(self):
+        headers = dict(Origin='localhost')
+        response = yield self.async_post('/', self.body, headers=headers)
+        self.assert_status_equal(json.loads(to_str(response.body)), 'Cross origin operation is not allowed.') # noqa
+
+    @tornado.testing.gen_test
+    def test_app_with_correct_event_origin(self):
+        body = dict(self.body, _origin=self.origin)
+        response = yield self.async_post('/', body)
+        self.assert_status_none(json.loads(to_str(response.body)))
+        self.assertIsNone(response.headers.get('Access-Control-Allow-Origin'))
+
+    @tornado.testing.gen_test
+    def test_app_with_correct_header_origin(self):
+        headers = dict(Origin=self.origin)
+        response = yield self.async_post('/', self.body, headers=headers)
+        self.assert_status_none(json.loads(to_str(response.body)))
+        self.assertEqual(
+            response.headers.get('Access-Control-Allow-Origin'), self.origin
+        )
