@@ -3,12 +3,13 @@ import json
 import logging
 import socket
 import struct
-import threading
 import traceback
 import weakref
 import paramiko
 import tornado.web
 
+from multiprocessing import cpu_count
+from concurrent.futures import ThreadPoolExecutor
 from tornado.ioloop import IOLoop
 from tornado.options import options
 from webssh.utils import (
@@ -17,11 +18,6 @@ from webssh.utils import (
 )
 from webssh.worker import Worker, recycle_worker, clients
 
-try:
-    from concurrent.futures import Future
-except ImportError:
-    from tornado.concurrent import Future
-
 try:
     from json.decoder import JSONDecodeError
 except ImportError:
@@ -173,6 +169,8 @@ class NotFoundHandler(MixinHandler, tornado.web.ErrorHandler):
 
 class IndexHandler(MixinHandler, tornado.web.RequestHandler):
 
+    executor = ThreadPoolExecutor(max_workers=cpu_count()*5)
+
     def initialize(self, loop, policy, host_keys_settings):
         super(IndexHandler, self).initialize(loop)
         self.policy = policy
@@ -275,10 +273,10 @@ class IndexHandler(MixinHandler, tornado.web.RequestHandler):
 
         if self.ssh_client._system_host_keys.lookup(key) is None:
             if self.ssh_client._host_keys.lookup(key) is None:
-                raise ValueError(
-                    'Connection to {}:{} is not allowed.'.format(
-                        hostname, port)
-                )
+                raise tornado.web.HTTPError(
+                        403, 'Connection to {}:{} is not allowed.'.format(
+                            hostname, port)
+                    )
 
     def get_args(self):
         hostname = self.get_hostname()
@@ -309,15 +307,9 @@ class IndexHandler(MixinHandler, tornado.web.RequestHandler):
 
         return result if result else 'utf-8'
 
-    def ssh_connect(self):
+    def ssh_connect(self, args):
         ssh = self.ssh_client
-
-        try:
-            args = self.get_args()
-        except InvalidValueError as exc:
-            raise tornado.web.HTTPError(400, str(exc))
-
-        dst_addr = (args[0], args[1])
+        dst_addr = args[:2]
         logging.info('Connecting to {}:{}'.format(*dst_addr))
 
         try:
@@ -337,14 +329,19 @@ class IndexHandler(MixinHandler, tornado.web.RequestHandler):
         worker.encoding = self.get_default_encoding(ssh)
         return worker
 
-    def ssh_connect_wrapped(self, future):
-        try:
-            worker = self.ssh_connect()
-        except Exception as exc:
-            logging.error(traceback.format_exc())
-            future.set_exception(exc)
-        else:
-            future.set_result(worker)
+    def check_origin(self):
+        event_origin = self.get_argument('_origin', u'')
+        header_origin = self.request.headers.get('Origin')
+        origin = event_origin or header_origin
+
+        if origin:
+            if not super(IndexHandler, self).check_origin(origin):
+                raise tornado.web.HTTPError(
+                    403, 'Cross origin operation is not allowed.'
+                )
+
+            if not event_origin and self.origin_policy != 'same':
+                self.set_header('Access-Control-Allow-Origin', origin)
 
     def head(self):
         pass
@@ -360,23 +357,21 @@ class IndexHandler(MixinHandler, tornado.web.RequestHandler):
 
         self.src_addr = self.get_client_addr()
         if len(clients.get(self.src_addr[0], {})) >= options.maxconn:
-            raise tornado.web.HTTPError(403, 'Too many connections.')
+            raise tornado.web.HTTPError(403, 'Too many live connections.')
 
-        origin = self.get_argument('_origin', u'')
-        if origin:
-            if not self.check_origin(origin):
-                raise tornado.web.HTTPError(
-                        403, 'Cross origin frame operation is not allowed.'
-                    )
+        self.check_origin()
 
-        future = Future()
-        t = threading.Thread(target=self.ssh_connect_wrapped, args=(future,))
-        t.setDaemon(True)
-        t.start()
+        try:
+            args = self.get_args()
+        except InvalidValueError as exc:
+            raise tornado.web.HTTPError(400, str(exc))
+
+        future = self.executor.submit(self.ssh_connect, args)
 
         try:
             worker = yield future
         except (ValueError, paramiko.SSHException) as exc:
+            logging.error(traceback.format_exc())
             self.result.update(status=str(exc))
         else:
             workers = clients.setdefault(worker.src_addr[0], {})
