@@ -6,24 +6,24 @@ from tornado.iostream import _ERRNO_CONNRESET
 from tornado.util import errno_from_exception
 
 
-BUF_SIZE = 1024
-workers = {}
+BUF_SIZE = 32 * 1024
+clients = {}
 
 
 def recycle_worker(worker):
     if worker.handler:
         return
     logging.warning('Recycling worker {}'.format(worker.id))
-    workers.pop(worker.id, None)
     worker.close(reason='worker recycled')
 
 
 class Worker(object):
-    def __init__(self, loop, ssh, chan, dst_addr):
+    def __init__(self, loop, ssh, chan, dst_addr, src_addr):
         self.loop = loop
         self.ssh = ssh
         self.chan = chan
         self.dst_addr = dst_addr
+        self.src_addr = src_addr
         self.fd = chan.fileno()
         self.id = str(id(self))
         self.data_to_dst = []
@@ -46,6 +46,8 @@ class Worker(object):
         if self.mode != mode:
             self.loop.update_handler(self.fd, mode)
             self.mode = mode
+        if mode == IOLoop.WRITE:
+            self.loop.call_later(0.1, self, self.fd, IOLoop.WRITE)
 
     def on_read(self):
         logging.debug('worker {} on read'.format(self.id))
@@ -94,11 +96,14 @@ class Worker(object):
 
     def close(self, reason=None):
         logging.info(
-            'Closing worker {} with reason {}'.format(self.id, reason)
+            'Closing worker {} with reason: {}'.format(self.id, reason)
         )
         if self.handler:
             self.loop.remove_handler(self.fd)
-            self.handler.close()
+            self.handler.close(reason=reason)
         self.chan.close()
         self.ssh.close()
         logging.info('Connection to {}:{} lost'.format(*self.dst_addr))
+
+        clients[self.src_addr[0]].pop(self.id, None)
+        logging.debug(clients)
